import einops
import torch
from torch import nn


class VitClassTokens(nn.Module):
    def __init__(self, dim: int, num_tokens: int = 1, location="first", init_std=0.02):
        super().__init__()
        self.location = location
        self.num_tokens = num_tokens
        self.init_std = init_std
        if num_tokens > 0:
            self.tokens = nn.Parameter(torch.zeros(1, num_tokens, dim))
        else:
            self.tokens = None
        self.reset_parameters()

    def reset_parameters(self):
        if self.num_tokens > 0:
            nn.init.normal_(self.tokens, std=self.init_std)

    def forward(self, x):
        if self.num_tokens == 0:
            return x
        assert x.ndim == 3
        tokens = self.tokens.expand(len(x), -1, -1)
        if self.location == "first":
            x = torch.concat([tokens, x], dim=1)
        elif self.location == "middle":
            pre, post = x.chunk(chunks=2, dim=1)
            x = torch.concat([pre, tokens, post], dim=1)
        else:
            raise NotImplementedError
        return x

    def split(self, x):
        if self.num_tokens == 0:
            return None, x
        assert x.ndim == 3
        if self.location == "first":
            cls_tokens = x[:, :self.num_tokens]
            patch_tokens = x[:, self.num_tokens:]
        else:
            raise NotImplementedError
        return cls_tokens, patch_tokens

    def pool(self, x):
        if self.num_tokens == 0:
            raise NotImplementedError
        if self.location == "first":
            x = einops.rearrange(x[:, :self.num_tokens], "batchsize seqlen dim -> batchsize (seqlen dim)")
        elif self.location == "middle":
            middle = x.size(1) // 2
            half_num_tokens = self.num_tokens // 2
            start = middle - half_num_tokens
            end = start + self.num_tokens
            x = x[:, start:end]
        else:
            raise NotImplementedError
        return x



from torch import nn


class Residual(nn.Sequential):
    """
    Wrapper for nn.Sequential for residual connections.
    Useful to include residual connections inside other sequentials
    Examples::
        >>> # manual residual connection
        >>> layer = nn.Linear(4, 4)
        >>> manual_y = x + layer(x)
        >>> # with Residual
        >>> layer = Residual(nn.Linear(4, 4))
        >>> residual_y = layer(x)

        >>> # block with 2 residual layers
        >>> layer0 = nn.Linear(4, 4)
        >>> layer1 = nn.Linear(4, 4)
        >>> manual_y = x + layer0(x)
        >>> manual_y = manual_y + layer1(manual_y)
        >>> # with Residual
        >>> block = nn.Sequential(Residual(nn.Linear(4, 4)), Residual(4, 4))
        >>> y = block(x)
    """

    def forward(self, x):
        return x + super().forward(x)



import torch
from torch import nn


class LayerScale(nn.Module):
    def __init__(self, dim: int, init_scale: float = 1e-5):
        super().__init__()
        self.gamma = nn.Parameter(torch.full(size=(dim,), fill_value=init_scale))

    def forward(self, x):
        return x * self.gamma



from functools import partial

from torch import nn

from kappamodules.attention import DotProductAttention1d
from kappamodules.init.functional import init_norms_as_noaffine
from kappamodules.layers import DropPath, LayerScale
from .mlp import Mlp


class PrenormBlock(nn.Module):
    def __init__(
            self,
            dim,
            num_heads,
            mlp_hidden_dim=None,
            qkv_bias=True,
            drop_path=0.,
            act_ctor=nn.GELU,
            norm_ctor=nn.LayerNorm,
            attn_ctor=DotProductAttention1d,
            layerscale=None,
            eps=1e-6,
            init_weights="xavier_uniform",
            init_norms="nonaffine",
            init_last_proj_zero=False,
    ):
        super().__init__()
        self.init_norms = init_norms
        mlp_hidden_dim = mlp_hidden_dim or dim * 4
        self.norm1 = norm_ctor(dim, eps=eps)
        self.attn = attn_ctor(
            dim=dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            init_weights=init_weights,
            init_last_proj_zero=init_last_proj_zero,
        )
        self.ls1 = nn.Identity() if layerscale is None else LayerScale(dim, init_scale=layerscale)
        self.drop_path1 = DropPath(drop_prob=drop_path)
        self.norm2 = norm_ctor(dim, eps=eps)
        self.mlp = Mlp(
            in_dim=dim,
            hidden_dim=mlp_hidden_dim,
            act_ctor=act_ctor,
            init_weights=init_weights,
            init_last_proj_zero=init_last_proj_zero,
        )
        self.ls2 = nn.Identity() if layerscale is None else LayerScale(dim, init_scale=layerscale)
        self.drop_path2 = DropPath(drop_prob=drop_path)
        self.reset_parameters()

    def reset_parameters(self):
        if self.init_norms == "torch":
            pass
        elif self.init_norms == "nonaffine":
            init_norms_as_noaffine(self.norm1)
            init_norms_as_noaffine(self.norm2)
        else:
            raise NotImplementedError

    def _attn_residual_path(self, x, attn_mask):
        return self.ls1(self.attn(self.norm1(x), attn_mask=attn_mask))

    def _mlp_residual_path(self, x):
        return self.ls2(self.mlp(self.norm2(x)))

    def forward(self, x, attn_mask=None):
        x = self.drop_path1(x, partial(self._attn_residual_path, attn_mask=attn_mask))
        x = self.drop_path2(x, self._mlp_residual_path)
        return x
